#!/usr/bin/perl
# ml_log_syncs.pl - Extract sync information from a v10 MobiLink server log.
#
# OOPERL_TYPE: ml

use strict;

#################################### TODO ######################
# TODO: Issue an error if a new instance starts in the middle of a file
#	    - Eventually handle this, somehow
# TODO: Handle CSV output when using both ACK and NoACK.
#	    --> Currently means some syncs don't correspond to captions
# TODO: Handle files from multiple ML servers in the same directory
#	    - Useful for global perspective on a server farm
#	    - Need to associate sync ids with servers, eg. <123(1)>
#################################### TODO ######################

#
# Debug.pm - Russ' core debugging class.
#
# Invoke the Dump methods as:
#   
#   Debug::Printf( level, str );

use strict;

#---------------------------------------------------------------------
{ package Debug;
#---------------------------------------------------------------------
    my $Level = 0;

    # Static Method
    sub Level {
	if( @_ ) {
	    $Level = shift();
	}
	return( $Level );
    }

    # Static
    sub Printf {
	my $level = shift();

	if( $level <= $Level ) {
	    my $fmt = shift() . "\n";
	    printf( $fmt, @_ );
	}
    }

    # Static
    sub PrintVar {
	my $level = shift();

	if( $Level > 0 ) {
	    my $var_name = shift();
	    Printf( $level, "$var_name = %s", eval( "\$$var_name" ) ); 
	}
    }

    # Static
    sub PrintArray {
	my $level = shift();

	if( $Level > 0 ) {
	    my $array_name = shift();
	    my $len = eval( "$array_name" );	# Get the array length
	    my $i;

	    for( $i = 0; $i < $len; $i++ ) {
		PrintVar( $level, "\$$array_name\[ $i \]" );
	    }
	}
    }
}

#
# Table.pm - Russ' core table-display classes.
#

use strict;

#require Obj;
#require Format_ASCII;

#---------------------------------------------------------------------
{ package Table;
#---------------------------------------------------------------------
    @Table::ISA = qw( Obj );

    my $num_tables = 0;

    sub new {
	my $proto	= shift();
	my $num_columns	= shift();
	my $border	= shift();
	if( $border eq '' ) {
	    $border = 1;
	}

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new();
	$num_tables++;
	if( $num_columns < 1 ) {
	    die( "Confused: Table defined with $num_columns columns." );
	}
	$self->{TABLE_NUM_COLUMNS} = $num_columns;
	$self->{TABLE_BORDER} = $border;
	$self->{TABLE_NUM_ROWS} = 0;
	$self->{TABLE_TEXT} = '';
	bless( $self, $class );

	Debug::Printf( 9, "New Table #$num_tables" );

	return( $self );
    };

    sub NumColumns {
	my $self = shift();
	return( $self->{TABLE_NUM_COLUMNS} );
    };

    sub Border {
	my $self = shift();
	return( $self->{TABLE_BORDER} );
    };

    sub NumRows {
	my $self = shift();
	return( $self->{TABLE_NUM_ROWS} );
    };

    sub Text {
	my $self = shift();
	return( $self->{TABLE_TEXT} );
    };

    sub IncRows {
	# Called directly by subclasses, and not the user.
	my $self = shift();
	my $num_columns = shift();

	if( $num_columns != $self->NumColumns ) {
	    die( "Expecting " . $self->NumColumns . " column values, but passed $num_columns" );
	}
	$self->{TABLE_NUM_ROWS} += 1;
    };

    sub Append {
	# The subclass is expected to provide an AddRow method
	# handling an array representing the row, and then calling Append().
	my $self = shift();
	$self->{TABLE_TEXT} .= shift();
    }

    sub DoneRows {
	# Subclass is expected to implement this method.
	die( "Confused: A table subclass isn't defining the DoneRows() method." );
    };
}

#---------------------------------------------------------------------
{ package ASCIITable;
#---------------------------------------------------------------------
    @ASCIITable::ISA = qw( Table );

    my $num_ascii_tables = 0;

    sub new {
	my $proto	= shift();
	my $num_columns	= shift();
	my $border	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( $num_columns, $border );
	$num_ascii_tables++;
	bless( $self, $class );

	$self->{ASCIIT_ROWS} = [];
	$self->{ASCIIT_CAPTION} = '';
	$self->{ASCIIT_COMMENT} = '';

	Debug::Printf( 9, "New ASCIITable #$num_ascii_tables" );

	return( $self );
    }

    sub Caption {
	my $self = shift();
	if( @_ ) {
	    my $caption = shift();
	    $self->{ASCIIT_CAPTION} = Format_ASCII::Underline( $caption );
	}
	return( $self->{ASCIIT_CAPTION} );
    }

    sub Comment {
	my $self = shift();
	if( @_ ) {
	    my $comment = shift();
	    $self->{ASCIIT_COMMENT} = "(comment)";
	}
	return( $self->{ASCIIT_COMMENT} );
    }

    sub AddRow {
	# The subclass is expected to provide an AddRow method
	# handling an array representing the row, and then calling Append().
	my $self = shift();
	my $rows = $self->{ASCIIT_ROWS};

	$self->IncRows( $#_ + 1 );
	
	# Defer formatting until all rows are seen.
	push( @$rows, [ @_ ] );
    }

    sub Width { #Static
	my @strs = split( /\n/, shift() );
	my $s;
	my $width = 0;
    
	# Return the width of the longest line.
	foreach $s (@strs) {
	    if( $width < length( $s ) ) {
		$width = length( $s );
	    }
	}
	return( $width );
    }

    sub Height { #Static
	my @strs = split( /\n/, shift() );
	return( $#strs + 1 );
    }

    sub GetSubLine { #Static
	my @strs = split( /\n/, shift() );
	my $index = shift() - 1;
Debug::Printf( 9, "Get index $index" );
	return( $strs[ $index ] );
    }

    sub DoneRows {
	my $self = shift();
	my $table = $self->{ASCIIT_ROWS};   # Array reference of array references (rows)
	my $row;
	my $v;
	my @widths;
	my $i;
	my $diff;
	my $type = 'asciiart';
	my $sep;
	my $sep2;
	my $h;

	for( $i = 0; $i < $self->NumColumns; $i++ ) {
	    push( @widths, 0 );
	}

	foreach $row (@$table) {
	    $i = 0;
	    Debug::Printf( 9, "ASCII Table Row: @$row" );
	    foreach $v (@$row) {
		if( $widths[ $i ] < Width( $v ) ) {
		    $widths[ $i ] = Width( $v );
		    Debug::Printf( 9, "\$widths[ $i ]: $widths[ $i ]" );
		}
		$i += 1;
	    }
	}

	# Format
	if( $self->Border ) {
	    foreach $i (@widths) {
		$sep .= '+' . '-' x ( $i + 2 );
	    }
	    $sep .= "+\n";
	    $sep2 = $sep;
	    $sep2 =~ s/-/=/g;
	    $self->Append( $sep );
	}

	foreach $row (@$table) {
	    my $height = 0;
	    # Calculate the line height.
	    foreach $v (@$row) {
		my $h = Height( $v );
		if( $height < $h ) {
		    $height = $h;
		}
	    }
	    Debug::Printf( 9, "Line height: $height" );

	    # Format the row.
	    for( $h = 1; $h <= $height; $h++ ) {
		my $str = '';
		$i = 0;
		foreach $v (@$row) {
		    my $xx;
		    $xx = GetSubLine( $v, $h );
		    Debug::Printf( 9, "Sub line $h: $xx" );
		    $diff = ( $widths[ $i ] - length( $xx ) );
		    Debug::Printf( 9, "Diff is $diff" );
		    if( $self->Border && $i == 0 ) {
			$str .= '| ';
		    }
		    $str .= $xx;
		    if( $diff > 0 ) {
			$str .= ( ' ' x $diff )
		    }
		    if( $self->Border ) {
			$str .= ' |';
		    }
		    if( $i < $#$row ) {
			$str .= ' ';
		    }
		    $i += 1;
		}

		$str .= "\n";
		$self->Append( $str );
	    }

	    if( $self->Border ) {
		if( $sep2 ne '' ) {
		    $self->Append( $sep2 );
		    $sep2 = '';
		} else {
		    $self->Append( $sep );
		}
	    }
	}
    }
}

#---------------------------------------------------------------------
{ package WikiTable;
#---------------------------------------------------------------------
    @WikiTable::ISA = qw( Table );

    my $num_wiki_tables = 0;

    sub new {
	my $proto	= shift();
	my $num_columns	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( $num_columns );
	$num_wiki_tables++;
	bless( $self, $class );

	$self->{WIKIT_CAPTION} = '';
	$self->{WIKIT_COMMENT} = '';

	Debug::Printf( 9, "New WikiTable #$num_wiki_tables" );

	return( $self );
    };

    sub Caption {
	my $self = shift();
	if( @_ ) {
	    my $caption = shift();
	    # Arbitrarily pick a level-two heading
	    $self->{WIKIT_CAPTION} = "---++ $caption";
	}
	return( $self->{WIKIT_CAPTION} );
    }

    sub Comment {
	my $self = shift();
	if( @_ ) {
	    my $comment = shift();
	    $self->{WIKIT_COMMENT} = "<!--\n$comment\n-->";
	}
	return( $self->{WIKIT_COMMENT} );
    }

    sub AddRow {
	# The subclass is expected to provide an AddRow method
	# handling an array representing the row, and then calling Append().
	my $self = shift();
	my @row = @_;
	my $c;
	my $str = '';
	my $header;
	my $ci;

	$self->IncRows( $#_ + 1 );
	$header = ( $self->NumRows == 1 );
	    
	$str .= ( $header ? '|* ' : '| ' );
	$ci = 0;
	foreach $c (@row) {
	    if( $ci++ > 0 ) {
		$str .= ( $header ? ' *|* ' : ' | ' );
	    }
	    $c =~ s/\n/<BR>/gm;
	    $str .= $c;
	}
	$str .= ( $header ? ' *|' : ' |' );
	$str .= "\n";
	$self->Append( $str );
    }

    sub DoneRows {
	# Nothing to do.
    };
}

#---------------------------------------------------------------------
{ package HTMLTable;
#---------------------------------------------------------------------
    @HTMLTable::ISA = qw( Table );

    # Colours
    my $GREEN    = '#00FF00';
    my $RED	 = '#FF0000';
    my $YELLOW	 = '#FFFF00';
    my $LT_GREY	 = '#EEEEEE';

    my $num_html_tables = 0;

    sub new {
	my $proto	= shift();
	my $num_columns	= shift();
	my $border	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( $num_columns );
	$num_html_tables++;
	bless( $self, $class );

	$self->{HTMLT_CAPTION} = '';
	$self->{HTMLT_COMMENT} = '';

	$self->Append( "<TABLE BORDER=$border CELLPADDING=3>\n" );

	Debug::Printf( 9, "New HTMLTable #$num_html_tables" );

	return( $self );
    };

    sub Caption {
	my $self = shift();
	if( @_ ) {
	    my $caption = shift();
	    # Arbitrarily pick a level-two heading
	    $self->{HTMLT_CAPTION} = "<H2> $caption </H2>";
	}
	return( $self->{HTMLT_CAPTION} );
    }

    sub Comment {
	my $self = shift();
	if( @_ ) {
	    my $comment = shift();
	    $self->{HTMLT_COMMENT} = "<!--\n$comment\n-->";
	}
	return( $self->{HTMLT_COMMENT} );
    }

    sub TextOnly { # static
	my $str = shift();

	# Remote tags. Strip leading and trailing whitespace.
	$str =~ s/<[^>]+>//gm;
	s/^\s+(.*)$/$1/;
	s/^(.*)\s+$/$1/;
	return( $str );
    }

    sub AddRow {
	# The subclass is expected to provide an AddRow method
	# handling an array representing the row, and then calling Append().
	my $self = shift();
	my @row = @_;
	my $c;
	my $str = '';
	my $header;

	$self->IncRows( $#_ + 1 );
	$header = ( $self->NumRows == 1 );
	    
	$str .= "<TR>\n";
	foreach $c (@row) {
	    if( $header ) {
		$str .= "<TH BGCOLOR=$LT_GREY>\n";
	    } else {
		my $text_only = TextOnly( $c );

		$str .= "<TD";
		# Format Yes and No as green and red, respectively.
		if( $text_only =~ /^y$/i ) {
		    $str .= " ALIGN=CENTER BGCOLOR=$GREEN";
		} elsif( $text_only =~ /^n$/i ) {
		    $str .= " ALIGN=CENTER BGCOLOR=$RED";
		} elsif( $text_only =~ /^p$/i ) {
		    $str .= " ALIGN=CENTER BGCOLOR=$YELLOW";
		}
		$str .= ">\n";
	    }
	    $c =~ s/\n/<BR>\n/gm;
	    $str .= $c;
	    $str .= ( $header ? "</TH>\n" : "</TD>\n" );
	}
	$str .= "</TR>\n";
	$self->Append( $str );
    }

    sub DoneRows {
	my $self = shift();
	$self->Append( "</TABLE>" );
    };
}

#
# Opts.pm - Russ' command-line options class.
#
use strict;

#use Obj;
#use Calendar;	    # Date
#use Table;	    # ASCIITable

my $OPT_TYPE_FLAG	= 'flag';
my $OPT_TYPE_UINT	= 'uint';
my $OPT_TYPE_STR	= 'str';
my $OPT_TYPE_ENUM_STR	= 'enumstr';
my $OPT_TYPE_DATE	= 'date';

#---------------------------------------------------------------------
{ package Opt;
#---------------------------------------------------------------------
    @Opt::ISA = qw( Obj );

    my $num_opt = 0;

    sub new {
	my $proto	= shift();
	my $type	= shift();
	my $default	= shift();
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new();
	$num_opt++;
	$self->{OPT_TYPE} = $type;
	$self->{OPT_DEFAULT} = $default;
	$self->{OPT_REQUIRED} = $required;
	$self->{OPT_OPTION} = $option;
	$self->{OPT_DESCRIPTION} = $description;
	$self->{OPT_VALUE} = $default;
	$self->{OPT_USED} = 0;
	bless( $self, $class );

	Debug::Printf( 9, "New Opt #$num_opt" );

	return( $self );
    };

    sub Type {
	my $self = shift();
	return( $self->{OPT_TYPE} );
    }

    sub Default {
	my $self = shift();
	# Allow the default to be set by subclasses
	if( @_ ) {
	    $self->{OPT_DEFAULT} = shift();
	}
	return( $self->{OPT_DEFAULT} );
    }

    sub Required {
	my $self = shift();
	return( $self->{OPT_REQUIRED} );
    }

    sub Option {
	my $self = shift();
	return( $self->{OPT_OPTION} );
    }

    sub Description {
	my $self = shift();
	return( $self->{OPT_DESCRIPTION} );
    }

    sub Value {
	my $self = shift();
	if( @_ ) {
	    $self->{OPT_VALUE} = shift();
	    $self->{OPT_USED} = 1;
	}
	return( $self->{OPT_VALUE} );
    }

    sub Used {
	my $self = shift();
	if( @_ ) {
	    # Back-door way of setting used.
	    # Useful for calling Value() in the constructor, then clearing it.
	    $self->{OPT_USED} = shift();
	}
	return( $self->{OPT_USED} );
    }

    sub IsAParameter {
	return( 0 );
    }

    sub IsAParameterSet {
	return( 0 );
    }
}

#---------------------------------------------------------------------
{ package OptFlag;
#---------------------------------------------------------------------
    @OptFlag::ISA = qw( Opt );

    my $num_opt_flags = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();

	my $class = ref($proto) || $proto;

	if( $default != 0 && $default != 1 ) {
	    die( "Confused about argument definition for flag $option." );
	}

	my $self = $class->SUPER::new( $OPT_TYPE_FLAG, $default, $required, $option, $description );
	$num_opt_flags++;
	bless( $self, $class );

	Debug::Printf( 9, "New OptFlag #$num_opt_flags" );

	return( $self );
    };
}

#---------------------------------------------------------------------
{ package OptParameter;
#---------------------------------------------------------------------
    @OptParameter::ISA = qw( Opt );

    my $num_opt_parameters = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'parameter', $default, $required, $option, $description );
	$num_opt_parameters += 1;
	bless( $self, $class );

	Debug::Printf( 9, "New OptParameter #$num_opt_parameters" );

	return( $self );
    };

    sub IsAParameter {
	return( 1 );
    }
}

#---------------------------------------------------------------------
{ package OptParameterUInt;
#---------------------------------------------------------------------
    @OptParameterUInt::ISA = qw( OptParameter );

    my $num_opt_parameter_uints = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();

	if( $default < 0 || $default !~ /^\d+$/ ) {
	    die( "The value $default is not an unsigned integer." );
	}

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( $default, $required, $option, $description );
	bless( $self, $class );

	Debug::Printf( 9, "New OptParameterUInt #$num_opt_parameter_uints" );

	return( $self );
    };

    sub Value {
	my $self = shift();
	if( @_ ) {
	    my $value = shift();
	    if( $value < 0 || $value !~ /^\d+$/ ) {
		die( "The value $value is not an unsigned integer." );
	    }
	    $self->SUPER::Value( $value );
	}
	return( $self->SUPER::Value );
    }
}

#---------------------------------------------------------------------
{ package OptParameterEnum;
#---------------------------------------------------------------------
    @OptParameterEnum::ISA = qw( OptParameter );

    my $num_opes = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();  # A list of enum values, comma-separated. First is default.
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();
	my @enum;

	my $class = ref($proto) || $proto;
	@enum = split( /,\s*/, $default );

	my $self = $class->SUPER::new( $enum[ 0 ], $required, $option, $description );
	$num_opes++;
	bless( $self, $class );

	$self->{PARAMENUMSTR_ENUM} = \@enum;

	# Set the default.
	if( $enum[ 0 ] eq '' ) {
	    die( "The default for OptEnum $option must be a non-empty, comma-separated list of enum values." );
	}
	Debug::Printf( 9, "Setting the default for option $option to $enum[0]" );
	$self->Default( $enum[ 0 ] );

	Debug::Printf( 9, "New OptParameterEnum #$num_opes" );

	return( $self );
    };

    sub Value {
	my $self = shift();
	if( @_ ) {
	    my $value = shift();
	    my $enum = $self->{PARAMENUMSTR_ENUM};

	    if( grep( /^$value$/i, @$enum ) ) {
		Debug::Printf( 9, "Setting option " . $self->Option . " to " . $value );
		$self->SUPER::Value( $value );
	    } else {
		die( "Illegal value for " . $self->Option . ": " . $value );
	    }
	}
	return( $self->SUPER::Value );
    }

    sub IsValue {
	my $self = shift();
	my $value = shift();

	return( $self->Value =~ /^$value$/i );
    }
}

#---------------------------------------------------------------------
{ package OptParameters;
#---------------------------------------------------------------------
    @OptParameters::ISA = qw( Opt );

    my $num_opt_parameterss = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'parameters', $default, $required, $option, $description );
	if( ++$num_opt_parameterss > 1 ) {
	    die( "Only one parameter set per program is allowed." );
	}
	bless( $self, $class );
	my $ref = ();

	# Cannot call Values(), since it sets used.
	my $values = $self->{ OPS_VALUE } = $ref;

	Debug::Printf( 9, "New OptParameters #$num_opt_parameterss" );

	return( $self );
    };

    sub IsAParameter {
	return( 1 );
    }

    sub IsAParameterSet {
	return( 1 );
    }

    # Append the value to the current array.
    sub Value {
	my $self = shift();
	if( @_ ) {
	    my $value = shift();
	    my $values = $self->{ OPS_VALUE };
	    $self->SUPER::Used( 1 );

	    push( @$values, $value ); $self->{ OPS_VALUE } = $values;
	}
	return( $self->{ OPS_VALUE } );
    }
}

#---------------------------------------------------------------------
{ package OptStr;
#---------------------------------------------------------------------
    @OptStr::ISA = qw( Opt );

    my $num_opt_str = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( $OPT_TYPE_STR, $default, $required, $option, $description );
	$num_opt_str++;
	bless( $self, $class );

	Debug::Printf( 9, "New OptStr #$num_opt_str" );

	return( $self );
    };
}

#---------------------------------------------------------------------
{ package OptEnum;
#---------------------------------------------------------------------
    @OptEnum::ISA = qw( Opt );

    my $num_opt_enum_str = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();  # A list of enum values, comma-separated. First is default.
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();
	my @enum;

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( $OPT_TYPE_ENUM_STR, $default, $required, $option, $description );
	$num_opt_enum_str++;
	bless( $self, $class );

	@enum = split( /,\s*/, $default );
	$self->{ENUMSTR_ENUM} = \@enum;
	$self->Value( $enum[0] );

	# Set the default.
	if( $enum[ 0 ] eq '' ) {
	    die( "The default for OptEnum $option must be a non-empty, comma-separated list of enum values." );
	}
	Debug::Printf( 9, "Setting the default for option $option to $enum[0]" );
	$self->Default( $enum[ 0 ] );
	$self->Used( 0 );

	Debug::Printf( 9, "New OptEnum #$num_opt_enum_str" );

	return( $self );
    };

    sub Value {
	my $self = shift();
	if( @_ ) {
	    my $value = shift();
	    my $enum = $self->{ENUMSTR_ENUM};

	    if( grep( /^$value$/i, @$enum ) ) {
		Debug::Printf( 9, "Setting option " . $self->Option . " to " . $value );
		$self->SUPER::Value( $value );
	    } else {
		die( "Illegal " . $self->Option . " enum value " . $value );
	    }
	}
	return( $self->SUPER::Value );
    }

    sub IsValue {
	my $self = shift();
	my $value = shift();

	return( $self->Value =~ /^$value$/i );
    }
}

#---------------------------------------------------------------------
{ package OptUInt;
#---------------------------------------------------------------------
    @OptUInt::ISA = qw( Opt );

    my $num_opt_uint = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( $OPT_TYPE_UINT, $default, $required, $option, $description );
	$num_opt_uint++;
	bless( $self, $class );

	Debug::Printf( 9, "New OptUInt #$num_opt_uint" );

	return( $self );
    };
}

#---------------------------------------------------------------------
{ package OptDate;
#---------------------------------------------------------------------
    @OptDate::ISA = qw( Opt );

    my $num_opt_date = 0;

    sub new {
	my $proto	= shift();
	my $default	= shift();
	my $required	= shift();
	my $option	= shift();
	my $description	= shift();
	my ( $year, $month, $day );

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( $OPT_TYPE_DATE, $default, $required, $option, $description );
	$num_opt_date++;
	bless( $self, $class );

	if( $default ne '' ) {
	    ( $year, $month, $day ) = split( /\//, $default );
	    $self->Value( $year, $month, $day );
	    $self->Used( 0 );
	}

	Debug::Printf( 9, "New OptDate #$num_opt_date" );

	return( $self );
    };

    sub Value {
	my $self = shift();

	if( @_ ) {  # year, month, day
	    my $date = new Date( shift(), shift(), shift() );
	    $self->{OPT_DATE_OBJ} = $date;
	}
	return( $self->{OPT_DATE_OBJ} );
    }
}

#---------------------------------------------------------------------
{ package Opts;
#---------------------------------------------------------------------
    @Opts::ISA = qw( Obj );

#    require Format_ASCII;   # Underline

    my $num_opts = 0;

    sub new {
	my $proto	= shift();
	my $program	= shift();
	my $argv	= shift();
	my $description = shift();
	my $notes       = shift();
	my @opts	= @_;
	my $a;
	my $o;

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new();
	if( ++$num_opts > 1 ) {
	    die( "There can only be one Opts instance per program" );
	}
	bless( $self, $class );

	$program =~ s/^.*\///;	# Remove the path.
	$program =~ s/^.*\\//;	# Remove the path.
	$self->{OPTS_PROGRAM}	  = $program;
	$self->{OPTS_DESCRIPTION} = $description;
	$self->{OPTS_NOTES}	  = $notes;
	$self->{OPTS_OPTS}	  = \@opts;

	Debug::Printf( 9, "New Opts #$num_opts" );

	while( $a = shift( @$argv ) ) {
	    my $found = 0;

	    if( $a =~ /^[-\/](.+)$/ ) {
		my $name = $1;

		O1: foreach $o (@opts) {
		    my $o_name = $o->Option;

		    if( $o->Type eq $OPT_TYPE_FLAG && $name =~ /^$o_name$/ ) {
			my $flag = !($o->Value);
			Debug::Printf( 9, "Toggling flag option " . $o->Option . " to: $flag" );
			$o->Value( $flag );
			$found = 1;
			last O1;
		    } elsif( $o->Type eq $OPT_TYPE_UINT && $name =~ /^$o_name=(\d+)$/ ) {
			Debug::Printf( 9, "Setting unsigned integer option " . $o->Option . " to: $1" );
			$o->Value( $1 );
			$found = 1;
			last O1;
		    } elsif( $o->Type eq $OPT_TYPE_DATE && $name =~ /^$o_name=((\d+)\/(\d+)\/(\d+))$/ ) {
			Debug::Printf( 9, "Setting date option " . $o->Option . " to: $1" );
			$o->Value( $2, $3, $4 );
			$found = 1;
			last O1;
		    } elsif( ( $o->Type eq $OPT_TYPE_STR || $o->Type eq $OPT_TYPE_ENUM_STR )
			  && $name =~ /^$o_name=(.+)$/ ) {
			Debug::Printf( 9, "Setting string option " . $o->Option . " to: $1" );
			$o->Value( $1 );
			$found = 1;
			last O1;
		    }
		}
		if( !$found ) {
		    $self->Usage( "Invalid option: $a" );
		}
	    } else {
		O2: foreach $o (@opts) {
		    if( $o->IsAParameter 
		     && ( $o->IsAParameterSet || !$o->Used ) ) {
			Debug::Printf( 9, "Setting parameter " . $o->Option . " to: $a" );
			$o->Value( $a );
			$found = 1;
			last O2;
		    }
		}
		if( !$found ) {
		    $self->Usage( "Invalid parameter: $a" );
		}
	    }
	}

	foreach $o (@opts) {
	    Debug::Printf( 1, "Option " . $o->Option . " Required=" . $o->Required . ", Used=" . $o->Used );
	    if( $o->Required && !$o->Used ) {
		if( $o->IsAParameter ) {
		    print( "Missing " . $o->Option . " parameter" );
		} else {
		    print( "Missing option: -" . $o->Option );
		}
		$self->Usage();
	    }
	}

	return( $self );
    };

    sub UsageFormat {
	my $self    = shift();
	my @output = @_;
	my $o;
	my $t = new ASCIITable( 3, 0 );

	# Indent and format.
	for( $o = 0; $o <= $#output; $o++ ) {
	    my @x = split( /\t/, $output[ $o ] );
	    $t->AddRow( $x[0], '  ', Format_ASCII::Wrap( $x[1], 50 ) );
	}
	$t->DoneRows();
	print( $t->Text . "\n" );
    }

    sub Usage
    {
	my $self = shift();
	my $reason = shift();
	my $desc = $self->{OPTS_DESCRIPTION};
	my $notes = $self->{OPTS_NOTES};
	my $program = $self->{OPTS_PROGRAM};
	my $opts = $self->{OPTS_OPTS};
	my $parms;
	my $last_parms;
	my $o;
	my $o_name;
	my @output;
	my $opt_parm_count;
	my $opt_last_parm_count;

	foreach $o (sort(@$opts)) {
	    if( $o->Type eq 'parameters' ) {
		$o_name = $o->Option;
		if( $o->Required ) {
		    $last_parms .= ']' x $opt_last_parm_count; $opt_last_parm_count = 0;
		}
		$last_parms .= ' ';
		if( $o->Required ) {
		    $last_parms .= "$o_name ...";
		} else {
		    $last_parms .= "[$o_name ...";
		    $opt_last_parm_count++;
		}
	    } elsif( $o->IsAParameter ) {
		$o_name = $o->Option;
		if( $o->Required ) {
		    $parms .= ']' x $opt_parm_count; $opt_parm_count = 0;
		}
		$parms .= ' ';
		if( $o->Required ) {
		    $parms .= "$o_name";
		} else {
		    $parms .= "[$o_name";
		    $opt_parm_count++;
		}
	    }
	}
	$parms .= ']' x $opt_parm_count; $opt_parm_count = 0;
	$parms .= $last_parms;
	$parms .= ']' x $opt_last_parm_count; $opt_last_parm_count = 0;
	if( @$opts ) {
	    $o = ' [options]';
	} else {
	    $o = '';
	}

	if( $reason ne '' ) {
	    print( "$reason\n" );
	}
	print( "\n" );
	print( Format_ASCII::Underline( 'Usage' ) . "\n" );
	print( "    $program$o$parms\n" );
	print( "\n" );
	print( "$desc\n" );
	print( "\n" );
	if( @$opts ) {
	    print( Format_ASCII::Underline( 'Options' ) . "\n" );

	    # Record the options.
	    foreach $o (@$opts) {
		if( !$o->IsAParameter ) {
		    my $desc = $o->Description;
		    my $str;
		    $o_name = $o->Option;

		    if( $o->Type eq $OPT_TYPE_STR
		     || $o->Type eq $OPT_TYPE_ENUM_STR ) {
			$str = "-$o_name=...";
		    } elsif( $o->Type eq $OPT_TYPE_UINT ) {
			$str = "-$o_name=N";
		    } elsif( $o->Type eq $OPT_TYPE_DATE ) {
			$str = "-$o_name=YYYY/MM/DD";
		    } else {
			$str = "-$o_name";
		    }
		    push( @output, "$str\t$desc" );
		}
	    }

	    $self->UsageFormat( @output );
	}

	if( $parms ) {
	    print( Format_ASCII::Underline( 'Parameters' ) . "\n" );

	    # Record the options.
	    while( @output ) {
		pop( @output );
	    }
	    foreach $o (@$opts) {
		if( $o->IsAParameter ) {
		    my $desc = $o->Description;
		    $o_name = $o->Option;

		    push( @output, "$o_name\t$desc" );
		}
	    }

	    $self->UsageFormat( @output );
	}

	if( $notes ne '' ) {
	    print( Format_ASCII::Underline( 'Notes' ) . "\n" );
	    print( "$notes\n" );
	}
	exit( -1 );
    }
}

#
# Obj.pm - Russ' core object class.
#

use strict;
#use Debug;

#---------------------------------------------------------------------
{ package Obj;
#---------------------------------------------------------------------
    my $num_objs = 0;

    sub new {
	my $proto = shift();

	my $class = ref($proto) || $proto;

	my $self = {};
	$num_objs++;
	bless( $self, $class );

	Debug::Printf( 9, '-------------------------' );
	Debug::Printf( 9, "New Obj #$num_objs" );

	return( $self );
    };
}

#
# dir.pm - Directory functions and objects.
#
#
{ package Dir;

use strict;
use Fcntl ':mode';
#-----------------------------------------------------------

#require Debug;
#require Platform;

#----------------------------------------------------------------
sub IsDir   # static
#----------------------------------------------------------------
{
    my $path = shift();
    my $is_dir;
    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
	$atime,$mtime,$ctime,$blksize,$blocks) = stat( $path );

    $is_dir =( S_ISDIR( $mode ) != 0 ); 
    Debug::Printf( 9, "ISDIR: $path\: $is_dir" );
    return( $is_dir );
}

#----------------------------------------------------------------
sub FileExists   # static
#----------------------------------------------------------------
{
    my $path = shift();
    my $file_exists;
    my @info = stat( $path );

    $file_exists = open( FE, "<$path" );
    if( $file_exists ) {
	close( FE );
    }
    Debug::Printf( 9, "FileExists $path\: $file_exists" );
    return( $file_exists );
}

#----------------------------------------------------------------
sub FileModTime   # static
#----------------------------------------------------------------
{
    my $path = shift();
    my $mod_time;
    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
	$atime,$mtime,$ctime,$blksize,$blocks) = stat( $path );

    $mod_time = localtime( $mtime );
    Debug::Printf( 9, "FileModTime $path\: $mod_time" );
    return( $mod_time );
}

my $MaxPathLen = 0;
my $MaxDisplay = 72;

#---------------------------------------------------------------------
sub DisplayCurrentPath
#---------------------------------------------------------------------
{
    my $path = shift();
    my $len  = length( $path );

    if( $MaxPathLen < $len ) {
	$MaxPathLen = $len;
    }
    if( $MaxPathLen > $MaxDisplay ) {
	$MaxPathLen = $MaxDisplay;
	$path = '...' . substr( $path, -( $MaxPathLen - 3 ) );
    }
    print( "\r" . ( ' ' x $MaxPathLen ) . "\r$path" );
}

#---------------------------------------------------------------------
{ package DirScanCallback;
#---------------------------------------------------------------------
    @DirScanCallback::ISA = qw( Obj );

    my $num_dirscancallback = 0;

    sub new {
	my $proto  = shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'DirScanCallback' );
	$num_dirscancallback++;
	bless( $self, $class );

	Debug::Printf( 9, "New DirScanCallback #$num_dirscancallback" );

	return( $self );
    };

    # Sub-class is expected to implement the Process method.
    sub Process {
	my $self = shift();
	my $path = shift();

	print( "$path\n" );
    }
}

#---------------------------------------------------------------------
{ package DirScan;
#---------------------------------------------------------------------
    @DirScan::ISA = qw( Obj );

    my $num_dirscan = 0;

    sub new {
	my $proto	= shift();
	my $root	= shift();
	my $dir_cb	= shift();
	my $file_cb	= shift();
	my $recurse	= shift();
	my @file_exts	= @_;

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'DirScan' );
	$num_dirscan++;
	bless( $self, $class );

	$self->{DIRSCAN_ROOT}	    = $root;
	$self->{DIRSCAN_DIR_CB}	    = $dir_cb;
	$self->{DIRSCAN_FILE_CB}    = $file_cb;
	$self->{DIRSCAN_RECURSE}    = $recurse;
	$self->{DIRSCAN_FILE_EXTS}  = \@file_exts;

	Debug::Printf( 9, "New DirScan #$num_dirscan" );

	return( $self );
    };

    sub Root {
	my $self = shift();
	return( $self->{DIRSCAN_ROOT} );
    }

    sub DirectoryCallback {
	my $self = shift();
	return( $self->{DIRSCAN_DIR_CB} );
    }

    sub FileCallback {
	my $self = shift();
	return( $self->{DIRSCAN_FILE_CB} );
    }

    sub Recurse {
	my $self = shift();
	return( $self->{DIRSCAN_RECURSE} );
    }

    sub FileExtensions {
	my $self = shift();
	my $fe = $self->{DIRSCAN_FILE_EXTS};

	return( @{$fe} );
    }

    sub ScanSubDir {	# Private
	my $self = shift();
	my $dir = shift();
	my $e;
	my $dir_cb  = $self->DirectoryCallback();
	my $file_cb = $self->FileCallback();
	my @dirs;
	
	Debug::Printf( 1, "Opening directory \"$dir\"" );
	opendir( ScanSubDir_D, $dir ) || die( "Unable to open directory $dir for read" );
	while( $e = readdir( ScanSubDir_D ) ) {
	    my $path = $dir . Platform::Slash() . $e;
	    if( Dir::IsDir( $path ) ) {
		if( $e !~ /^.*\.\.?$/ ) {
		    Debug::Printf( 2, "Found subdirectory \"$e\"" );
		    push( @dirs, $path );
		}
	    } else {
		if( $file_cb ne '' ) {
		    my @exts = $self->FileExtensions;

		    if( '' eq @exts ) {
			Debug::Printf( 2, "Found file \"$path\"" );
			$file_cb->Process( $path );
		    } else {
			my $e;

			FE: foreach $e (@exts) {
			    if( $path =~ /^.*$e$/ ) {
				Debug::Printf( 2, "Found file \"$path\"" );
				$file_cb->Process( $path );
				last FE;
			    }
			}
		    }
		}
	    }
	}
	Debug::Printf( 1, "Closing directory \"$dir\"" );
	closedir( ScanSubDir_D );

	foreach $e (sort(@dirs)) {
	    if( $dir_cb ne '' ) {
		$dir_cb->Process( $e );
	    }
	    if( $self->Recurse ) {
		$self->ScanSubDir( $e );
	    }
	}
    }

    sub Scan {
	my $self = shift();
    
	$self->ScanSubDir( $self->Root );
    }
}

#---------------------------------------------------------------------
{ package FileScanCallback;
#---------------------------------------------------------------------
    @FileScanCallback::ISA = qw( Obj );

    my $num_filescancallback = 0;

    sub new {
	my $proto  = shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'FileScanCallback' );
	$num_filescancallback++;
	bless( $self, $class );

	Debug::Printf( 9, "New FileScanCallback #$num_filescancallback" );

	return( $self );
    };

    # Sub-class is expected to implement the Process method.
    sub Process {
	my $self     = shift();
	my $file     = shift();
	my $line     = shift();
	my $line_num = shift();

	printf( "%s\n%08d\: %s\n", $file, $line_num, $line );
    }
}

#---------------------------------------------------------------------
{ package FileScan;
#---------------------------------------------------------------------
    @FileScan::ISA = qw( Obj );

    my $num_filescan = 0;

    sub new {
	my $proto	= shift();
	my $line_cb	= shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'FileScan' );
	$num_filescan++;
	bless( $self, $class );

	$self->{FILESCAN_LINE_CB} = $line_cb;

	Debug::Printf( 9, "New FileScan #$num_filescan" );

	return( $self );
    };

    sub LineCallback {
	my $self = shift();
	return( $self->{FILESCAN_LINE_CB} );
    }

    sub Scan {
	my $self	    = shift();
	my $file	    = shift();
	my $ok_if_not_open  = shift();
	my $e;
	my $line_cb = $self->LineCallback;
	
	Debug::Printf( 1, "Opening file \"$file\"" );
	if( open( Scan_F, $file ) ) {
	    while( <Scan_F> ) {
		chomp;
		if( $line_cb ne '' ) {
		    Debug::Printf( 9, "%08d\: %s\n", $., $_ );
		    $line_cb->Process( $file, $_, $. );
		}
	    }
	    Debug::Printf( 1, "Closing file \"$file\"" );
	    close( Scan_F );
	} elsif( $ok_if_not_open ) {
	    Debug::Printf( 1, "Ignoring missing file \"$file\"" );
	} else {
	    die( "Unable to open file $file for read" );
	}
    }
}

}

#
# format_ascii.pm - Simple routines for formatting ASCII text.
#

use strict;
#-----------------------------------------------------------

use Text::Wrap;
#require Obj;

{ package Format_ASCII;

    sub Underline {
	my $text = shift();
	my $bold = shift();
	my $char;

	if( $bold ) {
	    $char = '=';
	} else {
	    $char = '-';
	}
	return( "$text\n" . ( $char x length( $text ) ) );
    }

    sub Center {
	my $text  = shift();
	my $width = shift();
	my $len   = length( $text );
	my $pad   = int( ( $width - $len ) / 2 );
	if( $pad < 0 ) {
	    die( "Center: String of length $len cannot fit in width $width" );
	}
	$pad = ( ' ' x $pad );
	my $centered = $pad . $text . $pad;

	if( length( $centered ) != $width ) {
	    # Uneven padding
	    $centered .= ' ';
	}
	if( length( $centered ) != $width ) {
	    die( "Center: Unexpected length of " . length( $centered ) . ", expecting $width" );
	}
	return( $centered );
    }

    sub Wrap {
	if( $#_ + 1 != 2 ) {
	    die( "Invalid number of parameters to Format_ASCII::Wrap(). Expecting 2, but got " . ( $#_ + 1 ) );
	}
	my $text  = shift();
	my $width = shift();
	my $cols  = $Text::Wrap::columns;
	my $wrapped;

	$Text::Wrap::columns = $width;
	$wrapped = Text::Wrap::wrap( '', '', $text );
	$Text::Wrap::columns = $cols;
	return( $wrapped );
    }
}

#
# Platform.pm - Russ' platform-specific stuff.
#

use strict;

#---------------------------------------------------------------------
{ package Platform;
#---------------------------------------------------------------------
    my $IsWindows;
    my $Slash;
    my $OOPERL;

    $OOPERL = $ENV{ 'OOPERL' };
    if( $OOPERL eq '' ) {
	if( Platform::IsWindows() ) {
	    $OOPERL = "c:\\" . Platform::Slash() . 'ooperl';
	} else {
	    $OOPERL = $ENV{ 'HOME' } . Platform::Slash() . 'ooperl';
	}
    }

    if( $^O eq 'MSWin32' || $^O eq 'MSWin64' ) {
	$IsWindows = 1;
    } else {
	$IsWindows = 0;
    }

    if( $IsWindows ) {
	$Slash = "\\";
    } else {
	$Slash = '/';
    }

    sub IsWindows {
	return( $IsWindows );
    }

    sub Slash {
	return( $Slash );
    }

    sub MakePath {
        my $path = join Slash, @_;
        $path =~ s/\/\//\//g;

        return $path;
    }

    sub OOPERL {
	return( $OOPERL );
    }
}

#
# calendar.pm - Calendar objects.
#

use strict;
use Time::Local;
#-----------------------------------------------------------

#require Obj;
#require Opts;

#---------------------------------------------------------------------
{ package Month;
#---------------------------------------------------------------------
    @Month::ISA = qw( Obj );

    my $num_months = 0;

    sub new {
	my $proto	  = shift();
	my $name	  = shift();
	my $abbrev	  = shift();
	my $num_days	  = shift();
	my $first_weekday = shift();	# Origin 0

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'Month' );
	$num_months++;
	bless( $self, $class );

	Debug::Printf( 9, "New Month #$num_months" );

	$self->{MON_NAME}	    = $name;
	$self->{MON_ABBREV}	    = $abbrev;
	$self->{MON_NUM_DAYS}	    = $num_days;
	$self->{MON_FIRST_WEEKDAY}  = $first_weekday;

	return( $self );
    };

    sub Name {
	my $self = shift();
	return( $self->{MON_NAME} );
    }

    sub Abbrev {
	my $self = shift();
	return( $self->{MON_ABBREV} );
    }

    sub NumDays {
	my $self = shift();
	return( $self->{MON_NUM_DAYS} );
    }

    sub FirstWeekday {
	my $self = shift();
	return( $self->{MON_FIRST_WEEKDAY} );
    }

    sub Format {
	my $self = shift();
	my $wdy = $self->FirstWeekday;
	my $d;
	my @lines;
	my $l;

	$l = ' ' x ( $wdy * 3 );
	for( $d = 1; $d <= $self->NumDays; $d += 1 ) {
	    $l .= sprintf( "%3d", $d );
	    if( $wdy == 6 ) {
		push( @lines, $l );
		$l = '';
		$wdy = 0;
	    } else {
		$wdy += 1;
	    }
	}
	if( $wdy != 0 ) {
	    $l .= ' ' x ( ( 7 - $wdy ) * 3 );
	    push( @lines, $l );
	}
	return( @lines );
    }
}

#---------------------------------------------------------------------
{ package Year;
#---------------------------------------------------------------------
    @Year::ISA = qw( Obj );

    my $num_years = 0;

    sub new {
	my $proto  = shift();
	my $num    = shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'Year' );
	$num_years++;
	bless( $self, $class );

	Debug::Printf( 9, "New Year #$num_years" );

	$self->{YR_NUM} = $num;

	my $wdy;
	my $t = Time::Local::timegm( 0,0,0, 1, 0, ( $num - 1900 ) );
	my @lt = gmtime( $t );
	$wdy = $lt[ 6 ];

	# Add the months
	my $months = ();
	my @jan = ( 'January', 'Jan', 31, $wdy );
	$wdy += $jan[ 2 ]; $wdy = $wdy % 7;
	my @feb = ( 'February', 'Feb', 28, $wdy );
	if( $self->IsLeapYear ) {
	    $feb[2] += 1;
	}
	$wdy += $feb[ 2 ]; $wdy = $wdy % 7;
	my @mar = ( 'March', 'Mar', 31, $wdy );
	$wdy += $mar[ 2 ]; $wdy = $wdy % 7;
	my @apr = ( 'April', 'Apr', 30, $wdy );
	$wdy += $apr[ 2 ]; $wdy = $wdy % 7;
	my @may = ( 'May', 'May', 31, $wdy );
	$wdy += $may[ 2 ]; $wdy = $wdy % 7;
	my @jun = ( 'June', 'Jun', 30, $wdy );
	$wdy += $jun[ 2 ]; $wdy = $wdy % 7;
	my @jul = ( 'July', 'Jul', 31, $wdy );
	$wdy += $jul[ 2 ]; $wdy = $wdy % 7;
	my @aug = ( 'August', 'Aug', 31, $wdy );
	$wdy += $aug[ 2 ]; $wdy = $wdy % 7;
	my @sep = ( 'September', 'Sep', 30, $wdy );
	$wdy += $sep[ 2 ]; $wdy = $wdy % 7;
	my @oct = ( 'October', 'Oct', 31, $wdy );
	$wdy += $oct[ 2 ]; $wdy = $wdy % 7;
	my @nov = ( 'November', 'Nov', 30, $wdy );
	$wdy += $nov[ 2 ]; $wdy = $wdy % 7;
	my @dec = ( 'December', 'Dec', 31, $wdy );

	push( @$months, new Month( @jan ) );
	push( @$months, new Month( @feb ) );
	push( @$months, new Month( @mar ) );
	push( @$months, new Month( @apr ) );
	push( @$months, new Month( @may ) );
	push( @$months, new Month( @jun ) );
	push( @$months, new Month( @jul ) );
	push( @$months, new Month( @aug ) );
	push( @$months, new Month( @sep ) );
	push( @$months, new Month( @oct ) );
	push( @$months, new Month( @nov ) );
	push( @$months, new Month( @dec ) );

	$self->{YR_MONTHS} = $months;

	return( $self );
    };

    sub Number {
	my $self = shift();
	return( $self->{YR_NUM} );
    }

    sub IsLeapYear
    {
	my $self = shift();
	my $year = $self->{YR_NUM};

	if( int($year/4) == $year/4 ) {
	    if( int($year/100) == $year/100 ) {
		if( int($year/400) == $year/400 ) {
		    return( 1 );
		}
		return( 0 );
	    }
	    return( 1 );
	}
	return( 0 );
    }

    sub Months {
	my $self = shift();
	my $months = $self->{YR_MONTHS};
	return( @$months );
    }

    sub Month {
	my $self = shift();
	my $index = shift();
	my $months = $self->{YR_MONTHS};
	return( ${$months}[ $index ] );
    }
}

#---------------------------------------------------------------------
{ package Date;
#---------------------------------------------------------------------
    @Date::ISA = qw( Obj );

    my $num_dates = 0;

    sub new {
	my $proto  = shift();
	my $year   = shift();	# A number, empty ==> today
	my $month  = shift();	# A number, origin 1
	my $day    = shift();	# A number, origin 1

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'Date' );
	$num_dates++;
	bless( $self, $class );

	Debug::Printf( 9, "New Date #$num_dates" );

	if( $year eq '' ) {
	    my ($ltsec,$ltmin,$lthour,$ltmday,$ltmon,$ltyear,$ltwday,$ltyday,$ltisdst) = gmtime(time);
	    $year  = $ltyear + 1900;
	    $month = $ltmon + 1;
	    $day   = $ltmday;
	}
	$self->Set( $year, $month, $day );

	return( $self );
    };

    sub Set {
	my $self = shift();
	$self->{DATE_YEAR}  = shift();
	$self->{DATE_MONTH} = shift();
	$self->{DATE_DAY}   = shift();
	$self->{DATE_SSE}   = Time::Local::timegm( 0,0,0, $self->Day, $self->Month - 1, $self->Year - 1900 );


	my $t = $self->{DATE_SSE};
	my @lt = gmtime( $t );
	$self->{DATE_WEEKDAY} = $lt[ 6 ];
	Debug::Printf( 9, "Date->Set: wdy = %s", $self->WeekDay );
    }

    sub Year {
	my $self = shift();
	return( $self->{DATE_YEAR} );
    }

    sub Month {
	my $self = shift();
	return( $self->{DATE_MONTH} );
    }

    sub Day {
	my $self = shift();
	return( $self->{DATE_DAY} );
    }

    sub WeekDay {   # Origin 0
	my $self = shift();
	return( $self->{DATE_WEEKDAY} );
    }

    sub IsWeekDay {
	my $self = shift();
	my $wday = $self->WeekDay;
	return( $wday != 0 && $wday != 6 );
    }

    sub String {
	my $self = shift();
	return( sprintf( "%04d/%02d/%02d", $self->Year, $self->Month, $self->Day ) );
    }

    sub SecondsSinceEpoch {
	my $self = shift();
	return( $self->{DATE_SSE} );
    }

    sub AddDays {
	my $self = shift();
	my $num_days = shift();
	my $t = $self->SecondsSinceEpoch;

	Debug::Printf( 9, "AddDays: \$t = $t (before)" );
	$t += 60 * 60 * 24 * $num_days;
	Debug::Printf( 9, "AddDays: \$t = $t (after)" );
	my @lt = gmtime( $t );
	$self->Set( $lt[ 5 ] + 1900, $lt[ 4 ] + 1, $lt[ 3 ] );
    }
}

#---------------------------------------------------------------------
{ package DateTime;
#---------------------------------------------------------------------
    @DateTime::ISA = qw( Obj );

    my $num_datetimes = 0;

    sub new {
	my $proto  = shift();
	my $year   = shift();	# A number
	my $month  = shift();	# A number, origin 1
	my $day    = shift();	# A number, origin 1
	my $hour   = shift();	# A number, 0-23
	my $min    = shift();	# A number, 0-60
	my $sec    = shift();	# A number, 0-60

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'DateTime' );
	$num_datetimes++;
	bless( $self, $class );

	Debug::Printf( 9, "New DateTime #$num_datetimes" );

	$self->{DATETIME_YEAR}  = $year;
	$self->{DATETIME_MONTH} = $month;
	$self->{DATETIME_DAY}   = $day;
	$self->{DATETIME_HOUR}  = $hour;
	$self->{DATETIME_MIN}   = $min;
	$self->{DATETIME_SEC}   = $sec;

	$self->{DATETIME_SSE}   = Time::Local::timegm( $sec, $min, $hour, $day, $month - 1, $year - 1900 );

	return( $self );
    };

    sub Year {
	my $self = shift();
	return( $self->{DATETIME_YEAR} );
    }

    sub Month {
	my $self = shift();
	return( $self->{DATETIME_MONTH} );
    }

    sub Day {
	my $self = shift();
	return( $self->{DATETIME_DAY} );
    }

    sub Hour {
	my $self = shift();
	return( $self->{DATETIME_HOUR} );
    }

    sub Min {
	my $self = shift();
	return( $self->{DATETIME_MIN} );
    }

    sub Sec {
	my $self = shift();
	return( $self->{DATETIME_SEC} );
    }

    sub String {
	my $self = shift();
	return( sprintf( "%04d/%02d/%02d %02d:%02d:%02d",
		         $self->Year, $self->Month, $self->Day,
		         $self->Hour, $self->Min, $self->Sec ) );
    }

    sub SecondsSinceEpoch {
	my $self = shift();
	return( $self->{DATETIME_SSE} );
    }
}

#---------------------------------------------------------------------
sub MSToTime
#---------------------------------------------------------------------
{
    my $ms = shift();
    my $s = int( ( $ms + 500 ) / 1000 );
    my $h = int( $s / ( 60 * 60 ) );
    $s -= $h * 3600;
    my $m = int( $s / 60 );
    $s -= $m * 60;

    return( sprintf( "%02d:%02d:%02d", $h, $m, $s ) );
}

#---------------------------------------------------------------------
sub DateTimeFromStr
#---------------------------------------------------------------------
{
    my $str = shift();
    my $year;	# A number
    my $month;	# A number, origin 1
    my $day;	# A number, origin 1
    my $hour;	# A number, 0-23
    my $min;	# A number, 0-60
    my $sec;	# A number, 0-60

    $str =~ m/(\d\d\d\d)-(\d\d)-(\d\d) (\d\d):(\d\d):(\d\d)/;
    $year   = $1;
    $month  = $2;
    $day    = $3;
    $hour   = $4;
    $min    = $5;
    $sec    = $6;

    return( new DateTime( $year, $month, $day, $hour, $min, $sec ) );
}
#-----------------------------------------------------------

#require Obj;
#require Opts;
#require Platform;
#require Dir;
#require Calendar;
require IO::Handle;
STDOUT->autoflush(1);

my $Debug;
my $Filename;
my $OutputDirectory;
my $RemoveTimestamps;
my $RemoveSyncNums;
my $CSVFiles;
my $ShowProgress;
my $NoLogs;
my $IsDir;

my $DefaultTag = 'Main';
my %Syncs;
my %SyncLastLines;
my %SyncFirstLine;
my %SyncToType;
my %SyncToRemoteID;
my %SessionIDs;
my %TagMap;
my %SyncResult;
my %RemoteIDs;
my %ODBCTime;
my %TableODBCTime;
my %CopyTime;

my $TSPat = '\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d. ';

my $EVENT_PATTERN = 0;
my $EVENT_TYPE    = 1;
my @EventsNoACK = (
    [ 'Request from .+ for: remote ID:.*',		'request_from' ],
    [ 'Transaction: authenticate_user',			'authenticate_user' ],
    [ 'Transaction: begin_synchronization',		'begin_synchronization' ],
    [ 'Transaction: end_upload',			'end_upload' ],
    [ 'Transaction: prepare_for_download',		'prepare_for_download' ],
    [ 'Transaction: end_download',			'end_download' ],
    [ 'Transaction: end_synchronization',		'end_synchronization' ],
    [ 'Sending the download to the remote database.*',	'sending_download' ],
    [ 'Synchronization (complete|failed)',		'end_sync' ]
);
my $SendingDownloadPosNoACK = 9;

my @EventsACK = (
    [ 'Request from .+ for: remote ID:.*',		'request_from' ],
    [ 'Transaction: authenticate_user',			'authenticate_user' ],
    [ 'Transaction: begin_synchronization',		'begin_synchronization' ],
    [ 'Transaction: end_upload',			'end_upload' ],
    [ 'Transaction: prepare_for_download',		'prepare_for_download' ],
    [ 'Sending the download to the remote database.*',	'sending_download' ],
    [ 'Transaction: end_download',			'end_download' ],
    [ 'Transaction: end_synchronization',		'end_synchronization' ],
    [ 'Synchronization (complete|failed)',		'end_sync' ]
);
my $SendingDownloadPos  = 7;
my $EndDownloadPos	= 8;

my %SyncStart;		    # Key is sync id, value is DateTime object
my %SyncWorkStart;	    # Key is sync id, value is DateTime object
my %SyncEnd;		    # Key is sync id, value is DateTime object
my %SyncLastActivityTime;   # Key is sync id, value is DateTime object

my $FirstTimestamp;
my $LastTimestamp;

my %TimeCounts; # Key is timestamp, value is a unique @EventsACK-like array with counts based on @Events

my %EventTimes;	# Key is sync id (tag), value is a unique @EventsACK-like array with a DateTime per event,
                # except there is an extra entry at the end to indicate download ACK (1 ==> ACK)
my $FirstEventTime;
my $LastEventTime;

my $NumNoACKSyncs = 0;

#---------------------------------------------------------------------
sub GetEventCounts
#---------------------------------------------------------------------
{
    my $tm = shift();
    my $e;

    if( $TimeCounts{ $tm } eq '' ) {
	$TimeCounts{ $tm } = [];

	# Initialize counts to zero.
	foreach $e (@EventsACK) {
	    push( @{ $TimeCounts{ $tm } }, 0 );
	}
    }
    return( $TimeCounts{ $tm } );
}

#---------------------------------------------------------------------
sub GetEventTimes
#---------------------------------------------------------------------
{
    my $tag = shift();
    my $e;

    if( $EventTimes{ $tag } eq '' ) {
	$EventTimes{ $tag } = [];

	# Initialize times to zero (note these aren't DateTime objects).
	foreach $e (@EventsACK) {
	    push( @{ $EventTimes{ $tag } }, 0 );
	}

	# Add trailer indicating what array to use.
	push( @{ $EventTimes{ $tag } }, \@EventsACK );
    }
    return( $EventTimes{ $tag } );
}

#---------------------------------------------------------------------
sub CountEvent
#---------------------------------------------------------------------
{
    my $line = shift();
    my $tm   = shift();
    my $tag  = shift();
    my $type;
    my $counts;
    my $times;
    my $events_ref;
    my $pat;
    my $e;

    $counts = &GetEventCounts( $tm );
    $times  = &GetEventTimes( $tag );

    $events_ref = ${ $EventTimes{ $tag } }[ $#EventsACK + 1 ];

    # Keep track of download ACK (or not), as follows:
    #	- Wait until the end download message comes.
    #	- If the sending_download event isn't set yet, then set NoACK,
    #	  effectively reordering the events.
    if( $events_ref eq \@EventsACK ) {
	$pat = ${ $EventsACK[ $EndDownloadPos ] }[ $EVENT_PATTERN ];
	if( $line =~ m/^.*$pat$/ ) {
	    if( ${ $EventTimes{ $tag } }[ $SendingDownloadPos ] == 0 ) {
		Debug::Printf( 9, "<%s> Detected NO ACK", $tag );
		$events_ref = ${ $EventTimes{ $tag } }[ $#EventsACK + 1 ] = \@EventsNoACK;
		$NumNoACKSyncs += 1;
	    }
	}
    }

    for( $e = 0; $e <= $#{ $events_ref }; $e++ ) {
	$pat = ${ ${ $events_ref }[ $e ] }[ $EVENT_PATTERN ];
	if( $line =~ m/^.*$pat$/ ) {
	    if( $type eq 'sending_download' && $e != $SendingDownloadPos ) {
		# Switch from ACK sync to NO_ACK sync.
		${ $EventTimes{ $tag } } = \@EventsNoACK;
	    }

	    ${ $counts }[ $e ] += 1;
	    $type = ${ ${ $events_ref }[ $e ] }[ $EVENT_TYPE ];
	    Debug::Printf( 2, "%s %s = %s",
			   $tm, $type, ${ $counts }[ $e ] );

	    if( $FirstEventTime eq '' ) {
		$FirstEventTime = &DateTimeFromStr( $tm );
	    }
	    $LastEventTime = &DateTimeFromStr( $tm );

	    ${ $EventTimes{ $tag } }[ $e ] = &DateTimeFromStr( $tm );
	    last;
	}
    }
}

#---------------------------------------------------------------------
{ package MLLog;
#---------------------------------------------------------------------
    @MLLog::ISA = qw( Obj );

    my $num_mllog = 0;

    sub new {
	my $proto  = shift();
	my $path   = shift();
	my $is_dir = shift();

	my $class = ref($proto) || $proto;

	my $self = $class->SUPER::new( 'DirScanCallback' );
	$num_mllog++;
	bless( $self, $class );

	Debug::Printf( 9, "New MLLog #$num_mllog" );

	$self->{MLLOG_FIRST_FILE} = undef;
	$self->{MLLOG_PREV_FILES} = {};
	$self->{MLLOG_NEXT_FILES} = {};
	$self->{MLLOG_CURR_FILE}  = undef;
	$self->{MLLOG_CURR_FH}    = undef;
	$self->{MLLOG_PREV_TS}    = undef;

	if( $is_dir ) {
	    $self->AddFiles( $path );
	    $self->{MLLOG_DIR} = $path;
	} else {
	    $self->AddFile( undef, $path, undef );
	    $self->{MLLOG_DIR} = undef;
	}

	return( $self );
    };

    sub Dir {
	my $self = shift();

	return( $self->{MLLOG_DIR} );
    }

    sub FirstFile {
	my $self = shift();

	return( $self->{MLLOG_FIRST_FILE} );
    }

    sub CurrFile {
	my $self = shift();

	return( $self->{MLLOG_CURR_FILE} );
    }

    sub CurrFH {
	my $self = shift();

	return( $self->{MLLOG_CURR_FH} );
    }

    sub PrevFiles {
	my $self = shift();

	return( $self->{MLLOG_PREV_FILES} );
    }

    sub NextFiles {
	my $self = shift();

	return( $self->{MLLOG_NEXT_FILES} );
    }

    sub PrevFile {
	my $self = shift();
	my $file = shift();

	return( ${ $self->PrevFiles }{ $file } );
    }

    sub NextFile {
	my $self = shift();
	my $file = shift();

	return( ${ $self->NextFiles }{ $file } );
    }

    sub AddFile {
	my $self      = shift();
	my $dir       = shift();
	my $file      = shift();
	my $prev_file = shift();

	if( defined( $dir ) && !( -d $dir ) ) {
	    die( "$dir is not a directory" );
	}

	if( !( -f $prev_file ) ) {
	    $prev_file = undef;
	}

	if( defined( $prev_file ) ) {
	    Debug::Printf( 1, "Adding file %s, in dir %s, prev is %s", $file, $dir, $prev_file );
	} else {
	    Debug::Printf( 1, "Adding file %s, in dir %s", $file, $dir );
	}

	# Link files by pointing to the previous file.
	${ $self->PrevFiles }{ $file } = $prev_file;

	if( !defined( $self->FirstFile ) ) {
	    if( defined( $prev_file ) ) {
		$self->{MLLOG_FIRST_FILE} = $prev_file;
	    } else {
		$self->{MLLOG_FIRST_FILE} = $file;
	    }
	} else {
	    if( defined( $prev_file ) ) {
		if( !defined( ${ $self->PrevFiles }{ $prev_file } ) ) {
		    $self->{MLLOG_FIRST_FILE} = $prev_file; # Not found yet.
		}
		while( defined( ${ $self->PrevFiles }{ $self->FirstFile } ) ) {
		    $self->{MLLOG_FIRST_FILE} = ${ $self->PrevFiles }{ $self->FirstFile };
		}
	    } else {
		$self->{MLLOG_FIRST_FILE} = $file;
	    }
	}

	# Determine the next file(s).
	# Not efficient, but thorough.
	${ $self->NextFiles }{ $file } = undef;

	my $f;
	foreach $f (keys( %{ $self->PrevFiles } )) {
	    $prev_file = ${ $self->PrevFiles }{ $f };
	    if( defined( $prev_file ) ) {
		${ $self->NextFiles }{ $prev_file } = $f;
	    }
	}
    }

    sub AddFiles {
	my $self = shift();
	my $dir  = shift();
	my $e;

	opendir( D, $dir ) || die( "Unable to open directory $dir for read" );
	while( $e = readdir( D ) ) {
	    my $path = $dir . Platform::Slash() . $e;

	    if( !( -d $path ) ) {
		my $F;

		open( $F, "<$path" ) || die( "Unable to open $path for read" );
		my $first_line = readline( $F );
		close( $F );
		Debug::Printf( 2, "First line: %s", $first_line );

		if( $first_line =~ m/^[IWE]\. \d\d\d\d-.*/ ) {
		    if( $first_line =~ m/old output file "[^"]+" has been renamed to file "([^"]+)"/i ) {
			my $prev = lc( $1 );
			# Strip the directory, since it is redundant and prevents hash matches.
			$prev =~ s/.*\\(.*)/$1/;
			Debug::Printf( 2, "Prev: %s", $prev );
			# Add the file.
			$self->AddFile( $dir, $e, $prev );
		    } else {
			Debug::Printf( 1, "Ignoring %s", $path );
		    }
		} else {
		    Debug::Printf( 1, "Ignoring %s", $path );
		}
	    }
	}
	closedir( D );
    }

    sub OpenNextFile {
	my $self = shift();

	$self->{MLLOG_CURR_FH} = undef;

	if( !defined( $self->CurrFile ) ) {
	    $self->{MLLOG_CURR_FILE} = $self->FirstFile;
	} else {
	    $self->{MLLOG_CURR_FILE} = ${ $self->NextFiles }{ $self->CurrFile };
	}
	if( defined( $self->CurrFile ) ) {
	    my $f;
	    if( $self->Dir ne '' ) {
		$f = $self->Dir . Platform::Slash() . $self->CurrFile;
	    } else {
		$f = $self->CurrFile;
	    }
	    Debug::Printf( 1, "Opening %s...", $f );
	    open( $self->{MLLOG_CURR_FH}, "<$f" ) || die( "Unable to open $f for read" );
	} else {
	    $self->{MLLOG_CURR_FILE} = 'NO_MORE';
	}
    }

    sub ReadLine {
	my $self = shift();
	my $line;

	if( !defined( $self->CurrFile ) ) {
	    $self->OpenNextFile();
	}

	if( defined( $self->CurrFile ) && $self->CurrFile ne 'NO_MORE' ) {
	    $line = readline( $self->CurrFH );

	    if( !defined( $line ) ) {
		close( $self->CurrFH );
		if( defined( ${ $self->NextFiles }{ $self->CurrFile } ) ) {
		    $line = $self->{MLLOG_PREV_TS} . '<Main> * * * Extra, final line to simplify processing * * *';
		}
		$self->OpenNextFile();
		if( $self->CurrFile eq 'NO_MORE' ) {
		    $line = $self->{MLLOG_PREV_TS} . '<Main> * * * Extra, final line to simplify processing * * *';
		}
	    }
	}

	if( defined( $line ) ) {
	    if( $line =~ /^(.. )($TSPat)(<($DefaultTag|(\d+))> )(.*)$/ ) {
		$self->{MLLOG_PREV_TS} = "$1$2";
	    }
	}
	return( $line );
    }
}

#---------------------------------------------------------------------
sub GetSyncs
#---------------------------------------------------------------------
{
    my $file = shift();
    my $is_dir = shift();
    my $mll = new MLLog( $file, $is_dir );
    my $tag = '';
    my $line;

    while( $line = $mll->ReadLine() ) {
	my ( $s, $tm, $saved_tm, $sn, $c ) = ( '', '', '', '', '' );

	if( $ShowProgress->Value && ( $. % 1000 == 0 ) ) {
	    printf( "                                          \r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r%s line %d", $mll->CurrFile, $. );
	}

	chomp( $line );
	Debug::Printf( 5, "Line: %s", $line );
	if( $line =~ /^(.. )($TSPat)(<($DefaultTag|(\d+))> )(.*)$/ ) {
	    ( $s, $tm, $sn, $c ) = ( $1, $2, $3, $6 );
	    $saved_tm = $tm;
	    $tag = $4;

	    Debug::Printf( 6, "->Found a normal line" );

	    if( $tag ne $DefaultTag && $SyncResult{ $tag } eq '' ) {
		$SyncResult{ $tag } = 'unknown';
	    }
	    
	    if( $line =~ /^.*Request from "(.+)" for: remote ID:\s+(.+),\s+user.+$/i ) {
		Debug::Printf( 2, "Found request from %s for remote ID %s", $1, $2 );
		$SyncToType{ $tag } = $1;
		$SyncToRemoteID{ $tag } = $2;
		$RemoteIDs{ $2 } = 1;
	    } elsif( $line =~ /.*<(\d+)>.*new synchronization.*<(\d+)> with the same remote id..*/i ) {
		my ( $existing, $new ) = ( $1, $2 );
		$SyncToRemoteID{ $new } = $SyncToRemoteID{ $existing };
	    } elsif( $line =~ /.*<(\d+)>.*existing synchronization.*<(\d+)> with the same remote id..*/i ) {
		my ( $curr, $existing ) = ( $1, $2 );
		$SyncToRemoteID{ $curr } = $SyncToRemoteID{ $existing };
	    } elsif( $line =~ /.*Synchronization (complete|failed).*/ ) {
		$SyncResult{ $tag } = lc( $1 );
	    } elsif( $line =~ /.*ML timer.*/ ) {
		if( $line =~ /.*this table (\S+)\s+- used (\d+) ODBC time.*/ ) {
		    if( !defined( $TableODBCTime{ $s } ) ) {
			$TableODBCTime{ $tag } = {};
		    }
		    ${ $TableODBCTime{ $tag } }{ $1 } = $2;
		} elsif( $line =~ /.*download ended - used (\d+) ODBC time.*/ ) {
		    $ODBCTime{ $tag } = $1;
		} elsif( $line =~ /.*download ended - used (\d+) copy time.*/ ) {
		    $CopyTime{ $tag } = $1;
		} else {
		    # Ignore
		}
	    }

	    if( $RemoveTimestamps->Value ) {
		Debug::Printf( 6, "->Removing timestamps" );
		$tm = '';
	    }
	    if( $RemoveSyncNums->Value ) {
		Debug::Printf( 6, "->Removing sync numbers" );
		$sn = '';
	    }
	    $line = $s . $tm . $sn . $c;

	} elsif( $line =~ /^(.. )($TSPat)(.*)$/ ) {
	    ( $s, $tm, $c ) = ( $1, $2, $3 );
	    $saved_tm = $tm;

	    Debug::Printf( 6, "->Found a timestamped line without a tag" );
	    if( $RemoveTimestamps->Value ) {
		Debug::Printf( 6, "->Removing timestamps" );
		$tm = '';
	    }
	    if( $tag eq '' ) {
		Debug::Printf( 6, "->Removing sync numbers" );
		$tag = $DefaultTag;
	    }
	    $line = $s . $tm . $c;
	} else {
	    Debug::Printf( 6, "->Found a line without a timestamp" );
	    if( $tag eq '' ) {
		$tag = $DefaultTag;
	    }
	}

	if( $FirstTimestamp eq '' && $saved_tm ne '' ) {
	    $FirstTimestamp = $saved_tm;
	}
	if( $saved_tm ne '' ) {
	    $LastTimestamp = $saved_tm;
	}

	Debug::Printf( 6, "Modified Line: %s", $line );
	if( $TagMap{ $tag } ne '' ) {
	    # Special processing for non-persistent HTTP.
	    # Later connections get mapped onto an earlier connection.
	    Debug::Printf( 6, "->Tag <%s> maps to <%s>", $tag, $TagMap{ $tag } );
	    $tag = $TagMap{ $tag };
	}

	# Ignore empty lines, since they add no value,
	# and only complicate post-processing.
	if( $line !~ /^\s*$/ ) {
	    if( $NoLogs->Value ) {
		$Syncs{ $tag } = '';
	    } else {
		$Syncs{ $tag } .= $line . "\n";
	    }

	    # Record the first line of each sync.
	    if( !$NoLogs->Value ) {
		if( $SyncFirstLine{ $tag } eq '' ) {
		    $SyncFirstLine{ $tag } = "$line\n";
		}
	    }

	    # Record either the last info line (may span multiple lines),
	    # and/or the last contiguous set of warnings and/or errors.
	    if( $line =~ /^I\..*$/ ) {
		$SyncLastLines{ $tag } = "$line\n";
	    } else {
		$SyncLastLines{ $tag } .= "$line\n";
	    }
	}

	$saved_tm =~ s/\..+$//;
	if( $saved_tm ne '' ) {
	    if( $CSVFiles->Value ) {
		&CountEvent( $line, $saved_tm, $tag );
	    }

	    if( '' ne $tag && $tag =~ m/^\d+$/ ) {
		if( $SyncStart{ $tag } eq '' ) {
		    $SyncStart{ $tag } = $SyncEnd{ $tag } = $SyncLastActivityTime{ $tag } = DateTimeFromStr( $saved_tm );
		} else {
		    if( $CSVFiles->Value ) {
			if( $SyncWorkStart{ $tag } eq ''
			&& $line =~ m/^.*(COMMIT|ROLLBACK).*$/ ) {
			    $SyncWorkStart{ $tag } = DateTimeFromStr( $saved_tm );
			}
		    }
		    $SyncEnd{ $tag } = DateTimeFromStr( $saved_tm );
		    # Don't count the following error message as activity.
		    if( $line !~ /^.*new synchronization.*with the same remote id..*/i ) {
			$SyncLastActivityTime{ $tag } = $SyncEnd{ $tag };
		    }
		}
	    }
	}

	# Special processing for non-persistent HTTP
	if( $line =~ /^.*http: New session id: (.*)$/ ) {
	    Debug::Printf( 6, "->New Session id for <%s>: %s", $tag, $_ );
	    $SessionIDs{ $_ } = $tag;
	} elsif( $line =~ /^.*http: Continuing session for id: (.*)$/ ) {
	    # Attach the current sync number's lines to the original sync number.
	    my $actual_tag = $SessionIDs{ $_ };
	    Debug::Printf( 6, "->Mapping <%s> to <%s>", $tag, $actual_tag );
	    $Syncs{ $actual_tag } .= $Syncs{ $tag };
	    $Syncs{ $tag } = '';
	    $TagMap{ $tag } = $actual_tag;
	}
	Debug::Printf( 5, "-----------------------------------------------" );
    }

    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
sub ReportSyncs
#---------------------------------------------------------------------
{
    my $dir = shift();
    my $path;
    my $tag;
    my $count = 0;

    print( "Writing sync files...\n" );

    foreach $tag (keys(%Syncs)) {
	if( $Syncs{ $tag } ne '' ) {
	    $count += 1;
	    if( $ShowProgress->Value ) {
		print( "            \r\r\r\r\r\r\r\r\r\r\r\r$count" );
	    }
	    my $path = $dir . Platform::Slash() . 'sync_' . $tag . '.log';
	    Debug::Printf( 1, "Writing: %s", $path );
	    open( P, ">$path" ) || die( "Unable to open $path for write" );
	    print( P $Syncs{ $tag } );
	    close( P );
	}
    }
    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
sub ReportSyncsByRemoteID
#---------------------------------------------------------------------
{
    my $dir = shift();
    my $sep = '=' x 79;
    my $path;
    my $tag;
    my $count = 0;
    my $remote_id;
    my %got_rid;

    print( "Writing remote ID files...\n" );

    foreach $tag (sort( { $a <=> $b} keys(%Syncs))) {
	if( $Syncs{ $tag } ne '' ) {
	    $count += 1;
	    if( $ShowProgress->Value ) {
		print( "            \r\r\r\r\r\r\r\r\r\r\r\r$count" );
	    }

	    $remote_id = $SyncToRemoteID{ $tag };
	    if( $remote_id eq '' ) {
		$remote_id = 'UNKNOWN';
	    }

	    my $path = $dir . Platform::Slash() . 'rid_' . $remote_id . '.log';
	    if( !$got_rid{ $remote_id } ) {
		$got_rid{ $remote_id } = 1;
		Debug::Printf( 1, "Opening: %s", $path );
		open( P, ">$path" ) || die( "Unable to open $path for write" );
	    } else {
		Debug::Printf( 1, "Appending to: %s", $path );
		open( P, ">>$path" ) || die( "Unable to open $path for append" );
		print( P "$sep\n" );
	    }
	    print( P $Syncs{ $tag } );
	    close( P );
	}
    }
    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
sub ReportSyncLastLinesByRemoteID
#---------------------------------------------------------------------
{
    my $dir = shift();
    my $sep = '=' x 79;
    my $path;
    my $tag;
    my $count = 0;
    my $remote_id;
    my %got_rid;

    print( "Writing remote ID summary files...\n" );

    foreach $tag (sort( { $a <=> $b} keys(%Syncs))) {
	if( $Syncs{ $tag } ne '' ) {
	    $count += 1;
	    if( $ShowProgress->Value ) {
		print( "            \r\r\r\r\r\r\r\r\r\r\r\r$count" );
	    }

	    $remote_id = $SyncToRemoteID{ $tag };
	    if( $remote_id eq '' ) {
		$remote_id = 'UNKNOWN';
	    }

	    my $path = $dir . Platform::Slash() . 'rid_summary_' . $remote_id . '.log';
	    if( !$got_rid{ $remote_id } ) {
		$got_rid{ $remote_id } = 1;
		Debug::Printf( 1, "Opening: %s", $path );
		open( P, ">$path" ) || die( "Unable to open $path for write" );
	    } else {
		Debug::Printf( 1, "Appending to: %s", $path );
		open( P, ">>$path" ) || die( "Unable to open $path for append" );
		print( P "$sep\n" );
	    }
	    print( P $SyncFirstLine{ $tag } );
	    print( P $SyncLastLines{ $tag } );
	    close( P );
	}
    }
    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
sub ReportSummary
#---------------------------------------------------------------------
{
    my $dir = shift();
    my $sep = '=' x 79;
    my $path;
    my $tag;
    my $count = 0;
    my $remote_id;
    my %got_rid;
    my $path = $dir . Platform::Slash() . 'summary.txt';
    my @complete;
    my @failed;
    my @other;
    my $main_used = ( $Syncs{ $DefaultTag } ne '' ? 1 : 0 );
    my $num_syncs = 0 + (keys(%Syncs)) - $main_used;

    print( "Writing summary file: $path\...\n" );

    foreach $tag (sort( { $a <=> $b } keys(%SyncResult))) {
	$count += 1;
	if( $ShowProgress->Value ) {
	    print( "            \r\r\r\r\r\r\r\r\r\r\r\r$count" );
	}

	Debug::Printf( 9, "Sync <%s>: %s", $tag, $SyncResult{ $tag } );
	if( $SyncResult{ $tag } eq 'complete' ) {
	    push( @complete, $tag );
	} elsif( $SyncResult{ $tag } eq 'failed' ) {
	    push( @failed, $tag );
	} else {
	    push( @other, $tag );
	}
    }

    open( P, ">$path" ) || die( "Unable to open $path for write" );
    printf( P "# Remote IDs:  %6u\n", 0 + (keys(%RemoteIDs)) );
    printf( P "# syncs:       %6u\n", $num_syncs );
    if( $num_syncs == 0 ) {
	printf( P "  # complete:  0\n" );
	printf( P "  # failed:    0\n" );
	printf( P "  # other:     0\n" );
    } else {
	printf( P "  # complete:  %6u (%5.1f \%)\n", $#complete + 1, ( 1.0 + $#complete ) * 100.0 / $num_syncs );
	printf( P "  # failed:    %6u (%5.1f \%)\n", $#failed + 1, ( 1.0 + $#failed ) * 100.0 / $num_syncs );
	printf( P "  # other:     %6u (%5.1f \%)\n", $#other + 1, ( 1.0 + $#other ) * 100.0 / $num_syncs );
    }
    printf( P "\n" );
    printf( P "The following are the last lines of syncs that are neither complete nor failed,\n" );
    printf( P "ordered by time, enclosed by the first and last timstamped lines in the log.\n" );
    printf( P "Note that only the last _relevant_ line for a given sync is used for ordering\n" );
    printf( P "purposes. In other words, the last line showing sync activity always occurs\n" );
    printf( P "on or after the time of the last sync activity in the previous sync.\n" );
    printf( P "The purpose of all this is to order the last relevant activity in each sync\n" );
    printf( P "so we can hopefully highlight any blocking between concurrent syncs.\n" );
    printf( P "\n" );

    printf( P "-- %s <--- First timestamp in the log.\n", $FirstTimestamp );
    printf( P "$sep\n\n" );

    foreach $tag (sort( { my $ea = $SyncLastActivityTime{ $a }; my $eb = $SyncLastActivityTime{ $b }; return( $ea->SecondsSinceEpoch() <=> $eb->SecondsSinceEpoch() ); } @other)) {
	$remote_id = $SyncToRemoteID{ $tag };
	if( $remote_id eq '' ) {
	    $remote_id = 'UNKNOWN';
	}
	printf( P "Remote ID: $remote_id\n" );
	print( P $SyncLastLines{ $tag } );
	print( P "\n" );
    }

    printf( P "$sep\n" );
    printf( P "-- %s <--- Last timestamp in the log.\n", $LastTimestamp );

    close( P );

    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
sub ReportSyncTimes
#---------------------------------------------------------------------
{
    my $dir = shift();
    my $path;
    my $start;
    my $start_work;
    my $end;
    my $s;

    print( "Writing CSV file for sync times...\n" );

    my $path = $dir . Platform::Slash() . 'sync_times.csv';
    Debug::Printf( 1, "Writing: %s", $path );
    open( P, ">$path" ) || die( "Unable to open $path for write" );

    # Header
    print( P "Sync ID,Sync Work Time (s),Sync Total Time (s)\n" );
    foreach $s (sort( {$a <=> $b} keys(%SyncStart))) {
	$start	    = $SyncStart{ $s };
	$start_work = $SyncWorkStart{ $s };
	$end	    = $SyncEnd{ $s };
	if( $ShowProgress->Value ) {
	    print( "                              \r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r$s" );
	}
	if( $start ne '' && $start_work ne '' && $end ne '' ) {
	    Debug::Printf( 9, "sync_times.csv: <%s>: start = %s, start_work = %s, end = %s",
			   $s, $start->String, $start_work->String, $end->String );
	    printf( P "$s,%s,%s\n",
		    ( $end->SecondsSinceEpoch - $start_work->SecondsSinceEpoch ),
		    ( $end->SecondsSinceEpoch - $start->SecondsSinceEpoch ) );
	} else {
	    Debug::Printf( 9, "sync_times.csv: <%s>: Unable to determine CSV values" );
	}
    }

    close( P );

    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
sub ReportEvents
#---------------------------------------------------------------------
{
    my $dir = shift();
    my $path;
    my $e;
    my $type;
    my $count;
    my $tm;

    print( "Writing CSV file for events...\n" );

    my $path = $dir . Platform::Slash() . 'events.csv';
    Debug::Printf( 1, "Writing: %s", $path );
    open( P, ">$path" ) || die( "Unable to open $path for write" );

    # Header
    print( P "Timestamp" );
    for( $e = 0; $e <= $#EventsACK; $e++ ) {
	$type  = ${ $EventsACK[ $e ] }[ $EVENT_TYPE ];
	printf( P ",$type" );
    }
    print( P "\n" );

    foreach $tm (sort(keys(%TimeCounts))) {
	print( P "'$tm'" );
	for( $e = 0; $e <= $#EventsACK; $e++ ) {
	    $type  = ${ $EventsACK[ $e ] }[ $EVENT_TYPE ];
	    $count = ${ $TimeCounts{ $tm } }[ $e ];
	    if( $ShowProgress->Value ) {
		print( "                              \r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r$tm" );
	    }
	    printf( P ",%d", $count );
	}
	print( P "\n" );
    }

    close( P );

    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
sub ReportConcurrency
#---------------------------------------------------------------------
{
    my $dir = shift();
    my $path;
    my $e;
    my $type;
    my $tm;
    my $prev_tm;
    my $s_time;
    my $e_time;
    my $tag;
    my $secs;

    print( "Writing CSV file for concurrency...\n" );

    my $path = $dir . Platform::Slash() . 'concurrency.csv';
    Debug::Printf( 1, "Writing: %s", $path );
    open( P, ">$path" ) || die( "Unable to open $path for write" );

    # Header
    print( P "Timestamp,# Concurrent Syncs,# Syncs,# Listeners,# Others,# Monitors,Seconds Since Previous Entry\n" );

    foreach $tm (sort(keys(%TimeCounts))) {
	my $count = 0;
	my $sync_count = 0;
	my $dblsn_count = 0;
	my $monitor_count = 0;
	my $other_count = 0;

	if( $ShowProgress->Value ) {
	    print( "                              \r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r$tm" );
	}

	if( $prev_tm eq '' ) {
	    $secs = 0;
	} else {
	    Debug::Printf( 9, "tm: %s", $tm );
	    Debug::Printf( 9, "prev_tm: %s", $prev_tm );
	    my $prev = &DateTimeFromStr( $prev_tm );
	    my $curr = &DateTimeFromStr( $tm );
	    $secs = $curr->SecondsSinceEpoch - $prev->SecondsSinceEpoch;
	}
	$prev_tm = $tm;

	$tm =~ s/-/\//g;

	foreach $tag (keys(%SyncStart)) {
	    if( defined( $SyncStart{ $tag } ) ) {
		$s_time = $SyncStart{ $tag }->String;
	    } else {
		$s_time = '';
	    }
	    if( defined( $SyncEnd{ $tag } ) ) {
		$e_time = $SyncEnd{ $tag }->String;
	    } else {
		$e_time = '';
	    }
	    Debug::Printf( 9, "Comparing start, time, and end: %s : %s : %s", $s_time, $tm, $e_time );
	    if( $s_time ne '' && $e_time ne '' ) {
		if( $s_time le $tm && $e_time ge $tm ) {
		    Debug::Printf( 9, "(1) Sync %s active at %s", $tag, $tm );
		    $count++;
		    if( $SyncToType{ $tag } =~ m/^UL.*/i 
		     || $SyncToType{ $tag } =~ m/^dbmlsync.*/i ) {
			$sync_count++;
		    } elsif( $SyncToType{ $tag } =~ m/MLListener/i ) {
			$dblsn_count++;
		    } elsif( $SyncToType{ $tag } =~ m/MobiLink Monitor/i ) {
			$monitor_count++;
		    } else {
			$other_count++;
		    }
		}
	    } elsif( $s_time ne '' ) {
		if( $s_time le $tm ) {
		    Debug::Printf( 9, "(2) Sync %s active at %s", $tag, $tm );
		    $count++;
		    if( $SyncToType{ $tag } =~ m/^UL.*/i 
		     || $SyncToType{ $tag } =~ m/^dbmlsync.*/i ) {
			$sync_count++;
		    } elsif( $SyncToType{ $tag } =~ m/MLListener/i ) {
			$dblsn_count++;
		    } elsif( $SyncToType{ $tag } =~ m/MobiLink Monitor/i ) {
			$monitor_count++;
		    } else {
			$other_count++;
		    }
		}
	    } elsif( $e_time ne '' ) {
		if( $e_time ge $tm ) {
		    Debug::Printf( 9, "(3) Sync %s active at %s", $tag, $tm );
		    $count++;
		    if( $SyncToType{ $tag } =~ m/^UL.*/i 
		     || $SyncToType{ $tag } =~ m/^dbmlsync.*/i ) {
			$sync_count++;
		    } elsif( $SyncToType{ $tag } =~ m/MLListener/i ) {
			$dblsn_count++;
		    } elsif( $SyncToType{ $tag } =~ m/MobiLink Monitor/i ) {
			$monitor_count++;
		    } else {
			$other_count++;
		    }
		}
	    }
	}
	print( P "$tm,$count,$sync_count,$dblsn_count,$other_count,$monitor_count,$secs\n" );
    }

    close( P );

    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
sub ReportEventTimes
#---------------------------------------------------------------------
{
    my $dir = shift();
    my $path;
    my $e;
    my $start;
    my $end;
    my $count;
    my $tag;
    my $events_ref;
    my $ack_syncs = 0;
    my $no_ack_syncs = 0;

    print( "Writing CSV file for event times...\n" );

    my $path = $dir . Platform::Slash() . 'event_times.csv';
    Debug::Printf( 1, "Writing: %s", $path );
    open( P, ">$path" ) || die( "Unable to open $path for write" );

    # Header
    print( P "Sync ID" );
    foreach $tag (keys(%EventTimes)) {
	if( \@EventsACK == ${ $EventTimes{ $tag } }[ $#EventsACK + 1 ] ) {
	    $ack_syncs += 1;
	} else {
	    $no_ack_syncs += 1;
	}
    }
    if( $no_ack_syncs >= $ack_syncs ) {
	Debug::Printf( 1, "event_times.csv: Using NoACK headers" );
	$events_ref = \@EventsNoACK;
    } else {
	Debug::Printf( 1, "event_times.csv: Using ACK headers" );
	$events_ref = \@EventsACK;
    }
    for( $e = 0; $e <= $#{ $events_ref }; $e++ ) {
	if( $e == 0 ) {
	    $start = 'zero';
	    $end   = ${ ${ $events_ref }[ $e ] }[ $EVENT_TYPE ];
	} else {
	    $start = ${ ${ $events_ref }[ $e - 1 ] }[ $EVENT_TYPE ];
	    $end   = ${ ${ $events_ref }[ $e     ] }[ $EVENT_TYPE ];
	}
	printf( P ",$start->$end" );
    }
    print( P "\n" );

    foreach $tag (sort( {$a <=> $b} keys(%EventTimes))) {
	print( P "$tag" );
	if( $ShowProgress->Value ) {
	    print( "                              \r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r\r$tag" );
	}

	$events_ref = ${ $EventTimes{ $tag } }[ $#EventsACK + 1 ];
	if( $events_ref == \@EventsACK ) {
	    Debug::Printf( 9, "event_times.csv: <%s> ACK", $tag );
	} else {
	    Debug::Printf( 9, "event_times.csv: <%s> NO ACK", $tag );
	}

	for( $e = 0; $e <= $#{ $events_ref }; $e++ ) {
	    if( $e == 0 ) {
		$start = $FirstEventTime;
		$end   = ${ $EventTimes{ $tag } }[ $e ];
	    } else {
		$start = ${ $EventTimes{ $tag } }[ $e - 1 ];
		$end   = ${ $EventTimes{ $tag } }[ $e     ];
	    }
	    if( $start == 0 && $end != 0 ) {
		my $ee;
		# The start event didn't occur,
		# so find the last event that did occur.
		for( $ee = $e - 2; $ee >= 0; $ee-- ) {
		    $start = ${ $EventTimes{ $tag } }[ $ee ];
		    if( $start != 0 ) {
			Debug::Printf( 9, "event_times.csv: Scan back to previous event <%s> %s",
				    $tag, ${ ${ $events_ref }[ $ee ] }[ $EVENT_TYPE ] );
			last;
		    }
		}
		if( $start == 0 ) {
		    Debug::Printf( 9, "event_times.csv: Missing first event <%s> %s",
				   $tag, ${ ${ $events_ref }[ $e ] }[ $EVENT_TYPE ] );
		    $start = $FirstEventTime;
		} else {
		}
	    } elsif( $start != 0 && $end == 0 ) {
		# The end event didn't occur, so ignore it
		# unless it is the last possible event.
		if( $e == $#{ $events_ref } ) {
		    Debug::Printf( 9, "event_times.csv: Missing last event <%s> %s",
				   $tag, ${ ${ $events_ref }[ $e ] }[ $EVENT_TYPE ] );
		    $end = $LastEventTime;
		} else {
		    Debug::Printf( 9, "event_times.csv: Ignore end event <%s> %s",
				   $tag, ${ ${ $events_ref }[ $e ] }[ $EVENT_TYPE ] );
		    $start = 0;
		}
	    } elsif( $start == 0 && $end == 0 ) {
		Debug::Printf( 9, "event_times.csv: Ignoring event <%s> %s",
			       $tag, ${ ${ $events_ref }[ $e ] }[ $EVENT_TYPE ] );
		# Neither event occurred, so ignore this range.
	    } else {
		Debug::Printf( 9, "event_times.csv: Easy to calculate event time for <%s> %s",
			       $tag, ${ ${ $events_ref }[ $e ] }[ $EVENT_TYPE ] );
	    }

	    if( $start != 0 ) {
		my $time;

		Debug::Printf( 9, "event_times.csv: <%s> start = %s, end = %s (%s)",
			       $tag,
			       $start->String, $end->String,
			       ${ ${ $events_ref }[ $e ] }[ $EVENT_TYPE ] );
		$time = $end->SecondsSinceEpoch - $start->SecondsSinceEpoch;
		if( $time < 0 ) {
		    $time = 0;
		}
		printf( P ",%d", $time );
	    } else {
		print( P ',0' );
	    }
	}
	print( P "\n" );
    }

    close( P );

    if( $ShowProgress->Value ) {
	print( "\n" );
    }
}

#---------------------------------------------------------------------
# Mainline
#---------------------------------------------------------------------

# Default, Required, Option, Description
$Debug		 = new OptUInt( 0, 0, 'debug',
				'Debug level (0-9)' );
$Filename	 = new OptParameter( '', 1, 'filename',
				'The MobiLink server log file' );
$OutputDirectory = new OptParameter( '.', 0, 'directory',
				'Directory for the sync and csv files' );
$RemoveTimestamps= new OptFlag( 0, 0, 't',
				'Remove timestamps' );
$RemoveSyncNums  = new OptFlag( 0, 0, 's',
				'Remove sync numbers' );
$CSVFiles	 = new OptFlag( 0, 0, 'csv',
				'Generate *.csv files for important events' );
$ShowProgress	 = new OptFlag( 0, 0, 'p',
				'Show progress' );
$NoLogs		 = new OptFlag( 0, 0, 'nologs',
				'Do not generate log files' );
$IsDir		 = new OptFlag( 0, 0, 'd',
				'Input is a directory. Read all *.mls files' );
my $Opts = new Opts(
    $0, \@ARGV, 'Extract syncs from a v10 or v11 MobiLink server log.',
    "- Minimum required mlsrvNN verbosity is -v\n" .
    "- Output is to stdout\n" .
    "- The optional directory must already exist, if specified\n" .
    "- One summary file called summary.txt is created\n" .
    "- One log file is created per sync id, called sync_*.log\n" .
    "- One log file is created per remote id, called rid_*.log\n" .
    "- One summary file is created per remote id, called rid_summary_*.log\n" .
    "- CSV files are:\n" .
    "    - events.csv\n" .
    "    - event_times.csv\n" .
    "    - sync_times.csv\n" .
    "    - concurrency.csv"
    ,
    $Debug,
    $Filename,
    $OutputDirectory,
    $RemoveTimestamps,
    $RemoveSyncNums,
    $CSVFiles,
    $ShowProgress,
    $NoLogs,
    $IsDir
);
Debug::Level( $Debug->Value );

&GetSyncs( $Filename->Value, $IsDir->Value );
if( !$NoLogs->Value ) {
    &ReportSyncs( $OutputDirectory->Value );
    &ReportSyncsByRemoteID( $OutputDirectory->Value );
    &ReportSyncLastLinesByRemoteID( $OutputDirectory->Value );
}
&ReportSummary( $OutputDirectory->Value );
if( $CSVFiles->Value ) {
    &ReportSyncTimes( $OutputDirectory->Value );
    &ReportEvents( $OutputDirectory->Value );
    &ReportEventTimes( $OutputDirectory->Value );
    &ReportConcurrency( $OutputDirectory->Value );
}
print( "Done.\n" );
